Different directory markdown
<apply template="markdown"/>
This file doesn't have any actual test code referring to it because compiled
Heist will automatically fail the whole test suite if the case tested by this
template isn't tested correctly.
