<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<bind tag="att">ultralongname</bind>
<html>
<div id="pre_{att}_post">
/index
</div>
</html>
