<h:plain><h:use/> <h:use/></h:plain>
<h:defer><h:use/> <h:use/></h:defer>
<h:maydefer><h:use/> <h:use/></h:maydefer>
<h:maydefer2><h:use/> <h:use/></h:maydefer2>
<h:defermany><h:use/> <h:use/></h:defermany>
